@@ -11,7 +11,8 @@ import urllib3
 from pandas import json_normalize
 from nexus_api.NexusRequest import NexusRequest
 from nexus_api.NexusValue import NexusValue
-from nexus_api.exceptions import VoidDataframeException, CorruptDataframeException
+from nexus_api.exceptions import VoidDataframeException, CorruptDataframeException, NexusAPIException
+import warnings
 
 
 def WarningsAndJson(func):
@@ -20,7 +21,10 @@ def WarningsAndJson(func):
     def f(*args, **kwargs):
         urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
         rv = func(*args, **kwargs)
-        return rv.json()
+        if rv.status_code == 200:
+            return rv.json()
+        else:
+            raise NexusAPIException(rv)
 
     return f
 
@@ -50,8 +54,8 @@ def no_row_limit_decorator(fnc):
 
 
 class APINexus:
-    def __init__(self, IP_Maquina="localhost", Puerto=56000, token="", version="1.0"):
-        '''
+    def __init__(self, IP_Maquina="localhost", Puerto=56000, token="", version="1.0", logger=None):
+        """
         Metodo de inicialización de la clase. Aquí se define la IP y el puerto al que hemos de conectar así como el token
 
         Args:
@@ -59,8 +63,7 @@ class APINexus:
             Puerto (int): puerto de la API. 56000 de forma predeterminada
             token (str): token de la API
             version (str): 1.0 o 2.0 de momento. 1.0 by default
-        '''
-        print("Creada nueva instancia de tipo NEXUS API")
+        """
         self.IP_Maquina = IP_Maquina
         self.Puerto = str(Puerto)
         self.token = token
@@ -71,6 +74,34 @@ class APINexus:
             "nexusapiversion": self.version,
             "Content-Type": "application/json",
         }
+        self.logger = logger
+        self.log_print("Creada nueva instancia de tipo NEXUS API")
+
+    def log_print(self, message, severity='info'):
+        if self.logger is not None:
+            if severity=='info':
+                self.logger.info(message)
+            elif severity=='error':
+                self.logger.error(message)
+            elif severity=='debug':
+                self.logger.debug(message)
+            else:
+                raise ValueError('Severity not supported')
+        else:
+            print(message)
+
+
+    def __json_normalize_check(self, response) -> pd.DataFrame:
+        """
+        Intenta convertir a DataFrame la respuesta en json de la API. Si no es posible, es que ha habido un fallo (la respuesta no sigue el formato establecido)
+        Args:
+            response: json con valor de la llamada a NexusAPI
+        """
+        try:
+            return json_normalize(response)
+        except Exception as e:
+            self.log_print(f'Error de comunicación en NexusAPI. Motivo: {e}')
+            raise NexusAPIException(response)
 
     def statusConnection(self, url_completa):
         """COD respuesta de la API"""
@@ -182,24 +213,24 @@ class APINexus:
         variables = self.__getResponse(url_get)
         variables_names = list()
         try:
-            variables_norm = json_normalize(variables)
+            variables_norm = self.__json_normalize_check(variables)
             variables_names = list(variables_norm.name)
         except:
-            print(variables_names)
+            self.log_print(variables_names)
 
         if not variable_name in variables_names:
-            print("Se procede a crear la variable con nombre "
+            self.log_print("Se procede a crear la variable con nombre "
                   + variable_name
                   + ". Se devuelve json con uid y nombre de variable creada"
                   )
             payload = '["' + variable_name + '"]'
-            print(payload)
+            self.log_print(payload)
             url_post = self.url_NX + "/api/Tags/insert"
-            print(url_post)
+            self.log_print(url_post)
             response = self.__postResponse(url_post, payload)
-            dataReceived = json_normalize(response)
+            dataReceived = self.__json_normalize_check(response)
         else:
-            print(
+            self.log_print(
                 "La variable ya existe, no se creará ninguna variable. Se devuelve el json con el uid de la variable existente"
             )
             dataReceived = variables_norm[variables_norm.name == variable_name]
@@ -218,17 +249,17 @@ class APINexus:
         """
         if uid_tagview is None:
             tagviews = self.callGetDocuments()
-            tagviews = json_normalize(tagviews)
+            tagviews = self.__json_normalize_check(tagviews)
             uid_tagview = tagviews.uid[0]
         vbles = pd.DataFrame(self.callGetTagViews(uid_tagview))['columns']
-        vbles = json_normalize(vbles)
+        vbles = self.__json_normalize_check(vbles)
         if filters is None:
             uids = vbles['uid'].to_list()
         else:
             filters = [filters] if isinstance(filters, str) else filters
             uids = vbles[vbles.name.isin(filters)].uid.to_list()
         # Read RT values from tagview
-        data = json_normalize(self.callGetTagViewsRealTime(uid_tagview, uids))
+        data = self.__json_normalize_check(self.callGetTagViewsRealTime(uid_tagview, uids))
         diccio = dict([(i, j) for i, j in zip(vbles.uid, vbles.name)])
         data['timeStamp'] = pd.to_datetime(data['timeStamp'], unit='s')
         data['name'] = data['uid'].map(diccio)
@@ -242,9 +273,9 @@ class APINexus:
         """
         # La función comprueba primero si existe la variable en la que se quiere escribir
         url = self.url_NX + "/api/Tags/writable"
-        # print(url_completa)
+        # self.log_print(url_completa)
         variables = self.__getResponse(url)
-        variables_norm = json_normalize(variables)
+        variables_norm = self.__json_normalize_check(variables)
         variables_names = list(variables_norm.name)
 
         # La función escribe en la variable
@@ -262,7 +293,7 @@ class APINexus:
             )
                     + "]"
             )
-            # print(payload)
+            # self.log_print(payload)
             # payload= "[{\"uid\": \"" + variable_uid + "\",\"value\": " + str(variable_value) + ",\"timeStamp\": " + str(timestamp) + "}]"
 
             headers = {
@@ -284,13 +315,54 @@ class APINexus:
                         + variable_uid
                         + " pero la operación ha fallado"
                 )
-                print(response.text.encode("utf8"))
+                self.log_print(response.text.encode("utf8"))
         # Si no existe, devuelve un mensaje para que el usuario cree la variable deseada (no lo hace automático para evitar creación de variables por errores tipográficos
         else:
             dataReceived = "La variable en la que se ha solicitado escribir no existe. Por favor creela mediante la función callPostTagInsert"
 
         return dataReceived
 
+    def callPostValue_operate(self, variable_uid, variable_value):
+        """Escritura de variable en tiempo real.
+        args:
+            variable_uid: nombre de la variable a escribir en el PLC
+            variable_value: valor de la variable a escribir
+        """
+
+        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
+        url_completa = self.url_NX + "/api/Tags/operate"
+        nexusvalue = NexusValue(variable_uid, variable_value)
+        payload = (
+                "["
+                + json.dumps(
+            nexusvalue, default=lambda o: o.__dict__, sort_keys=False, indent=2
+        )
+                + "]"
+        )
+
+        headers = {
+            "nexustoken": self.token,
+            "nexusapiversion": self.version,
+            "Content-Type": "application/json",
+        }
+        response = requests.request(
+            "POST", url_completa, headers=headers, data=payload
+        )
+
+        if response.status_code == 200:
+            dataReceived = "Escritura correcta"
+        else:
+            dataReceived = (
+                    "Se ha intentado escribir en la variable "
+                    + " con uid "
+                    + variable_uid
+                    + " pero la operación ha fallado"
+            )
+        self.log_print(response.text.encode("utf8"))
+        # Si no existe, devuelve un mensaje para que el usuario cree la variable deseada (no lo hace automático para evitar creación de variables por errores tipográficos
+
+        return dataReceived
+
     @no_row_limit_decorator
     def callPostValueHist(self, df_value_timestamp):
         """Función de escritura de variable para diferentes historicos.
@@ -300,14 +372,14 @@ class APINexus:
         # la funcion mira cuantas variables diferentes contiene el dataframe y comprueba si todas ellas existen
         vbles = list(df_value_timestamp.name.unique())
         n = len(vbles)
-        print('se intenta escribir en ' + str(n) + ' variables')
+        self.log_print('se intenta escribir en ' + str(n) + ' variables')
         # La función comprueba primero si existe la variable en la que se quiere escribir
         urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
         url_completa = self.url_NX + "/api/Tags/writable"
         response = requests.get(url_completa, verify=False, headers=self.header)
         variables = response.json()
         variables_pd = pandas.DataFrame(variables)
-        variables_norm = json_normalize(variables)
+        variables_norm = self.__json_normalize_check(variables)
         diccio = dict([(i, j) for i, j in zip(variables_pd.name, variables_pd.uid)])
 
         variables_names = list(variables_norm.name)
@@ -315,7 +387,7 @@ class APINexus:
         for j in vbles:
             if j not in variables_names:
                 NOK = 1
-                print('la variable ' + str(j) + 'no esta creada ')
+                self.log_print('la variable ' + str(j) + 'no esta creada ')
 
         if NOK == 0:
             df2 = df_value_timestamp.copy()
@@ -323,9 +395,9 @@ class APINexus:
 
             df2.drop(columns=["name"], inplace=True)
             if df2['timeStamp'].dtype == 'datetime64[ns]':
-                print('timeStamp in datetime format: please check that it is in UTC')
+                warnings.warn('timeStamp in datetime format: please check that it is in UTC')
                 df2['timeStamp'] = df2['timeStamp'].astype('int64') / 1e9
-            print(str(df2))
+            self.log_print(str(df2))
             payload = pandas.DataFrame.to_json(
                 df2, date_format="epoch", orient="records"
             )
@@ -347,7 +419,7 @@ class APINexus:
                         "Se ha intentado escribir en la variable "
                         + " pero la operación ha fallado" + str(response.status_code)
                 )
-                print(response.text.encode("utf8"))
+                self.log_print(response.text.encode("utf8"))
         # Si no existe, devuelve un mensaje para que el usuario cree la variable deseada (no lo hace automático para evitar creación de variables por errores tipográficos
         else:
             dataReceived = "La variable en la que se ha solicitado escribir no existe. Por favor creela mediante la función callPostTagInsert"
@@ -365,14 +437,14 @@ class APINexus:
         # la funcion mira cuantas variables diferentes contiene el dataframe y comprueba si todas ellas existen
         vbles = list(df_value_timestamp.name.unique())
         n = len(vbles)
-        print('se intenta escribir en ' + str(n) + ' variables')
+        self.log_print('se intenta escribir en ' + str(n) + ' variables')
         # La función comprueba primero si existe la variable en la que se quiere escribir
         urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
         url_completa = self.url_NX + "/api/Tags/writable"
         response = requests.get(url_completa, verify=False, headers=self.header)
         variables = response.json()
         variables_pd = pandas.DataFrame(variables)
-        variables_norm = json_normalize(variables)
+        variables_norm = self.__json_normalize_check(variables)
         diccio = dict([(i, j) for i, j in zip(variables_pd.name, variables_pd.uid)])
 
         variables_names = list(variables_norm.name)
@@ -380,7 +452,7 @@ class APINexus:
         for j in vbles:
             if j not in variables_names:
                 NOK = 1
-                print('la variable ' + str(j) + 'no esta creada ')
+                self.log_print('la variable ' + str(j) + 'no esta creada ')
 
         if NOK == 0:
             df2 = df_value_timestamp.copy()
@@ -388,9 +460,9 @@ class APINexus:
 
             df2.drop(columns=["name"], inplace=True)
             if df2['timeStamp'].dtype == 'datetime64[ns]':
-                print('timeStamp in datetime format: please check that it is in UTC')
+                warnings.warn('timeStamp in datetime format: please check that it is in UTC')
                 df2['timeStamp'] = df2['timeStamp'].astype('int64') / 1e9
-            print(str(df2))
+            self.log_print(str(df2))
             payload = pandas.DataFrame.to_json(
                 df2, date_format="epoch", orient="records"
             )
@@ -412,7 +484,7 @@ class APINexus:
                         "Se ha intentado escribir en la variable "
                         + " pero la operación ha fallado" + str(response.status_code)
                 )
-                print(response.text.encode("utf8"))
+                self.log_print(response.text.encode("utf8"))
         # Si no existe, devuelve un mensaje para que el usuario cree la variable deseada (no lo hace automático para evitar creación de variables por errores tipográficos
         else:
             dataReceived = "La variable en la que se ha solicitado escribir no existe. Por favor creela mediante la función callPostTagInsert"
@@ -432,7 +504,7 @@ class APINexus:
         url = self.url_NX + "/api/Tags/writable"
         variables = self.__getResponse(url)
         variables_pd = pandas.DataFrame(variables)
-        variables_norm = json_normalize(variables)
+        variables_norm = self.__json_normalize_check(variables)
         diccio = dict([(i, j) for i, j in zip(variables_pd.name, variables_pd.uid)])
         df2 = df_variable_name_value
         df2['uid'] = df2['name'].map(diccio)
@@ -444,11 +516,11 @@ class APINexus:
         for j in vbles:
             if j not in variables_names:
                 NOK = 1
-                print('la variable ' + str(j) + 'no esta creada')
+                self.log_print('la variable ' + str(j) + 'no esta creada')
 
         if NOK == 0:
             # La función escribe en la variable
-            print('se actualiza el valor RT de ' + str(n) + ' vbles')
+            self.log_print('se actualiza el valor RT de ' + str(n) + ' vbles')
             urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
             url_completa = self.url_NX + "/api/Tags/realtime/insert"
             headers = {
@@ -464,11 +536,12 @@ class APINexus:
                         "Se ha intentado escribir en la variable "
                         + " pero la operación ha fallado" + str(response.status_code)
                 )
-                print(response.text.encode("utf8"))
+                self.log_print(response.text.encode("utf8"))
 
         return dataReceived
 
-    def filter_installation(self, Datefrom, Dateto, columnas, filter_txt=None, uids=None, resolucion=3, fuente=0) -> pd.DataFrame:
+    def filter_installation(self, Datefrom, Dateto, columnas, filter_txt=None, uids=None, resolucion=3,
+                            fuente=0) -> pd.DataFrame:
         """
         la funcion recibe como parametros la fecha ini, fecha fin, un df con los uid y
         los nombres de las variables de la instalación y el filtro de texto aplicar
@@ -497,7 +570,7 @@ class APINexus:
                 if filter_txt:
                     uids = list(columnas[columnas['name'].str.contains(filter_txt, case=False)].uid)
             if not uids:
-                print('Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
+                self.log_print('Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
                 uids = list(columnas.uid)
         # Remove duplicate UIDS
         uids = list(set(uids))
@@ -505,7 +578,7 @@ class APINexus:
         fecha_fin = Dateto
         nexusRequest = NexusRequest(uids, fecha_ini, fecha_fin, fuente, resolucion)
         filtered_hist = self.callGetTagsHistory(nexusRequest)
-        filtered_hist = json_normalize(filtered_hist)
+        filtered_hist = self.__json_normalize_check(filtered_hist)
         # Check that there is no error in the dataframe
         self.__check_response_error_df(filtered_hist)
         filtered_hist.timeStamp = pandas.to_datetime(filtered_hist.timeStamp, unit='s')
@@ -513,7 +586,8 @@ class APINexus:
         filtered_hist['name'] = filtered_hist['uid'].map(diccio)
         return filtered_hist
 
-    def filter_tagview(self, Datefrom, Dateto, columnas, uid_vista, filter_txt=None, uids=None, resolucion=3, fuente=0) -> pd.DataFrame:
+    def filter_tagview(self, Datefrom, Dateto, columnas, uid_vista, filter_txt=None, uids=None, resolucion=3,
+                       fuente=0) -> pd.DataFrame:
         """
         La funcion recibe como parametros la fecha ini, fecha fin, un df con los uid y
         los nombres de las variables de la instalación y el filtro de texto aplicar [pueden ser varios]
@@ -542,7 +616,7 @@ class APINexus:
                 if filter_txt:
                     uids = list(columnas[columnas['name'].str.contains(filter_txt, case=False)].uid)
             if not uids:
-                print('Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
+                self.log_print('Los filtros proporcionados no encuentran ninguna variable. Se devolverá toda la instalación')
                 uids = list(columnas.uid)
         # Remove duplicate UIDS
         uids = list(set(uids))
@@ -553,7 +627,7 @@ class APINexus:
         # RES_5_MIN, RES_200_MIL, RES_500_MIL, RES_1_SEC ]
         nexusRequest = NexusRequest(uids, fecha_ini, fecha_fin, fuente, resolucion)
         filtered_hist = self.callGetDataviewHistory(uid_vista, nexusRequest)
-        filtered_hist = json_normalize(filtered_hist)
+        filtered_hist = self.__json_normalize_check(filtered_hist)
         # Check dataframe for errors
         self.__check_response_error_df(filtered_hist)
         filtered_hist.timeStamp = pandas.to_datetime(filtered_hist.timeStamp, unit='s')
@@ -562,13 +636,13 @@ class APINexus:
         return filtered_hist
 
     def get_alarms_uids_by_names(self, names):
-        alarms = json_normalize(self.callGetAlarms())
+        alarms = self.__json_normalize_check(self.callGetAlarms())
         names = [names] if isinstance(names, str) else names
         uids = alarms['uid'][alarms['name'].isin(names)]
         return uids
 
     def get_alarms_uids_by_groups(self, groups):
-        alarms = json_normalize(self.callGetAlarms())
+        alarms = self.__json_normalize_check(self.callGetAlarms())
         groups = [groups] if isinstance(groups, str) else groups
         uids = []
         for group in groups:
